@@ -495,7 +495,11 @@ ar_member_touch (arname, memname)
   if (AR_HDR_SIZE != write (fd, (char *) &ar_hdr, AR_HDR_SIZE))
     goto lose;
   /* The file's mtime is the time we we want.  */
+#ifdef EINTR
+  while (fstat (fd, &statbuf) < 0 && errno == EINTR);
+#else
   fstat (fd, &statbuf);
+#endif
 #if defined(ARFMAG) || defined(AIAMAG)
   /* Advance member's time to that time */
   for (i = 0; i < sizeof ar_hdr.ar_date; i++)
