@@ -21,6 +21,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "job.h"
 #include "file.h"
 #include "variable.h"
+#include <assert.h>
 
 /* Default path to search for executables.  */
 static char default_path[] = ":/bin:/usr/bin";
@@ -28,6 +29,16 @@ static char default_path[] = ":/bin:/usr/bin";
 /* Default shell to use.  */
 char default_shell[] = "/bin/sh";
 
+#ifdef __MSDOS__
+#include <process.h>
+static int dos_pid = 123;
+static int dos_status;
+static char *dos_bname;
+static char *dos_bename;
+static int dos_batch_file;
+#endif /* MSDOS.  */
+
+
 /* If NGROUPS_MAX == 0 then try other methods for finding a real value.  */
 #if defined (NGROUPS_MAX) && NGROUPS_MAX == 0
 #undef NGROUPS_MAX
@@ -260,6 +271,7 @@ reap_children (block, err)
 	}
       else if (pid == 0)
 	{
+#ifndef	__MSDOS__
 	  /* No remote children.  Check for local children.  */
 
 	  if (any_local)
@@ -308,6 +320,14 @@ reap_children (block, err)
 	      exit_sig = WIFSIGNALED (status) ? WTERMSIG (status) : 0;
 	      coredump = WCOREDUMP (status);
 	    }
+#else	/* MSDOS.  */
+	  /* Life is very different on MSDOS.  */
+	  pid = dos_pid - 1;
+	  status = dos_status;
+	  exit_code = dos_status;
+	  exit_sig = 0;
+	  coredump = 0;
+#endif	/* Not MSDOS.  */
 	}
       else
 	/* We got a remote child.  */
@@ -359,9 +379,15 @@ reap_children (block, err)
 	    {
 	      /* The commands failed.  Write an error message,
 		 delete non-precious targets, and abort.  */
+	      static int delete_on_error = -1;
 	      child_error (c->file->name, exit_code, exit_sig, coredump, 0);
 	      c->file->update_status = 1;
-	      if (exit_sig != 0)
+	      if (delete_on_error == -1)
+		{
+		  struct file *f = lookup_file (".DELETE_ON_ERROR");
+		  delete_on_error = f != 0 && f->is_target;
+		}
+	      if (exit_sig != 0 || delete_on_error)
 		delete_child_targets (c);
 	    }
 	  else
@@ -383,7 +409,6 @@ reap_children (block, err)
 			 Since there are more commands that wanted to be run,
 			 the target was not completely remade.  So we treat
 			 this as if a command had failed.  */
-		      c->file->command_state = cs_finished;
 		      c->file->update_status = 1;
 		    }
 		  else
@@ -394,32 +419,33 @@ reap_children (block, err)
 			 by start_remote_job_p.  */
 		      c->remote = start_remote_job_p ();
 		      start_job_command (c);
+		      if (c->file->command_state == cs_running)
+			/* We successfully started the new command.
+			   Loop to reap more children.  */
+			continue;
 		    }
-		}
 
-	      switch (c->file->command_state)
-		{
-		case cs_running:
-		  /* Successfully started.  Loop to reap more children.  */
-		  continue;
-
-		case cs_finished:
 		  if (c->file->update_status != 0)
 		    /* We failed to start the commands.  */
 		    delete_child_targets (c);
-		  break;
-
-		default:
-		  error ("internal error: `%s' has bogus command_state \
-%d in reap_children",
-			 c->file->name, (int) c->file->command_state);
-		  abort ();
-		  break;
 		}
+	      else
+		/* There are no more commands.  We got through them all
+		   without an unignored error.  Now the target has been
+		   successfully updated.  */
+		c->file->update_status = 0;
 	    }
 
+	  /* When we get here, all the commands for C->file are finished
+	     (or aborted) and C->file->update_status contains 0 or 1.  But
+	     C->file->command_state is still cs_running if all the commands
+	     ran; notice_finish_file looks for cs_running to tell it that
+	     it's interesting to check the file's modtime again now.  */
+
 	  if (! handling_fatal_signal)
-	    /* Notice if the target of the commands has been changed.  */
+	    /* Notice if the target of the commands has been changed.
+	       This also propagates its values for command_state and
+	       update_status to its also_make files.  */
 	    notice_finished_file (c->file);
 
 	  if (debug_flag)
@@ -475,6 +501,13 @@ free_child (child)
 }
 
 #ifdef	POSIX
+#ifdef	__MSDOS__
+void
+unblock_sigs ()
+{
+  return;
+}
+#else
 extern sigset_t fatal_signal_set;
 
 void
@@ -485,6 +518,7 @@ unblock_sigs ()
   sigprocmask (SIG_SETMASK, &empty, (sigset_t *) 0);
 }
 #endif
+#endif
 
 /* Start a job to run the commands specified in CHILD.
    CHILD is updated to reflect the commands and ID of the child process.  */
@@ -495,9 +529,14 @@ start_job_command (child)
 {
   static int bad_stdin = -1;
   register char *p;
-  int flags = child->file->cmds->lines_flags[child->command_line - 1];
+  int flags;
   char **argv;
 
+  /* Combine the flags parsed for the line itself with
+     the flags specified globally for this target.  */
+  flags = (child->file->command_flags
+	   | child->file->cmds->lines_flags[child->command_line - 1]);
+
   p = child->command_ptr;
   child->noerror = flags & COMMANDS_NOERROR;
   while (*p != '\0')
@@ -519,7 +558,7 @@ start_job_command (child)
   if (question_flag && !(flags & COMMANDS_RECURSE))
     {  
       child->file->update_status = 1;
-      child->file->command_state = cs_finished;
+      notice_finished_file (child->file);
       return;
     }
 
@@ -555,6 +594,12 @@ start_job_command (child)
       /* This line has no commands.  Go to the next.  */
       if (job_next_command (child))
 	start_job_command (child);
+      else
+	{
+	  /* No more commands.  All done.  */
+	  child->file->update_status = 0;
+	  notice_finished_file (child->file);
+	}
       return;
     }
 
@@ -580,6 +625,7 @@ start_job_command (child)
       free ((char *) argv);
       if (job_next_command (child))
 	start_job_command (child);
+      child->file->update_status = 0;
       return;
     }
 
@@ -618,6 +664,8 @@ start_job_command (child)
   if (child->environment == 0)
     child->environment = target_environment (child->file);
 
+#ifndef	__MSDOS__
+
   /* start_waiting_job has set CHILD->remote if we can start a remote job.  */
   if (child->remote)
     {
@@ -667,10 +715,26 @@ start_job_command (child)
 	}
     }
 
+#else	/* MSDOS.  */
+  dos_status = spawnvpe (P_WAIT, argv[0], argv, child->environment);
+  ++dead_children;
+  child->pid = dos_pid++;
+  if (dos_batch_file)
+   {
+     dos_batch_file = 0;
+     remove (dos_bname);	/* Ignore errors.  */
+     if (access (dos_bename, 0))
+       dos_status = 1;
+     else
+       dos_status = 0;
+     remove (dos_bename);
+   }
+#endif	/* Not MSDOS.  */
+
   /* We are the parent side.  Set the state to
      say the commands are running and return.  */
 
-  child->file->command_state = cs_running;
+  set_command_state (child->file, cs_running);
 
   /* Free the storage used by the child's argument list.  */
 
@@ -681,7 +745,7 @@ start_job_command (child)
 
  error:
   child->file->update_status = 2;
-  child->file->command_state = cs_finished;
+  notice_finished_file (child->file);
 }
 
 /* Try to start a child running.
@@ -727,15 +791,18 @@ start_waiting_job (c)
       unblock_sigs ();
       break;
 
+    case cs_not_started:
+      /* All the command lines turned out to be empty.  */
+      c->file->update_status = 0;
+      /* FALLTHROUGH */
+
     case cs_finished:
       notice_finished_file (c->file);
       free_child (c);
       break;
 
     default:
-      error ("internal error: `%s' command_state == %d in new_job",
-	     c->file->name, (int) c->file->command_state);
-      abort ();
+      assert (c->file->command_state == cs_finished);
       break;
     }
 
@@ -881,8 +948,11 @@ new_job (file)
 
   /* Fetch the first command line to be run.  */
   if (! job_next_command (c))
-    /* There were no commands!  */
-    free_child (c);
+    {
+      /* There were no commands!  */
+      free_child (c);
+      c->file->update_status = 0;
+    }
   else
     {
       /* The job is now primed.  Start it running.  */
@@ -910,8 +980,6 @@ job_next_command (child)
 	{
 	  /* There are no more lines to be expanded.  */
 	  child->command_ptr = 0;
-	  child->file->command_state = cs_finished;
-	  child->file->update_status = 0;
 	  return 0;
 	}
       else
@@ -924,6 +992,9 @@ job_next_command (child)
 static int
 load_too_high ()
 {
+#ifdef	__MSDOS__
+  return 1;
+#else
   extern int getloadavg ();
   double load;
 
@@ -949,6 +1020,7 @@ load_too_high ()
   user_access ();
 
   return load >= max_load_average;
+#endif
 }
 
 /* Start jobs that are waiting for the load to be lower.  */
@@ -1012,7 +1084,13 @@ search_path (file, path, program)
   if (path == 0 || path[0] == '\0')
     path = default_path;
 
-  if (index (file, '/') != 0)
+  if (
+#ifdef __MSDOS__
+      strpbrk (file, "/\\:")
+#else
+      index (file, '/') 
+#endif
+      != 0)
     {
       strcpy (program, file);
       return 1;
@@ -1049,7 +1127,7 @@ search_path (file, path, program)
 	  int perm;
 	  char *p;
 
-	  p = index (path, ':');
+	  p = index (path, PATH_SEPARATOR_CHAR);
 	  if (p == 0)
 	    p = path + strlen (path);
 
@@ -1062,7 +1140,7 @@ search_path (file, path, program)
 	      bcopy (file, program + (p - path) + 1, len);
 	    }
 
-	  if (stat (program, &st) == 0
+	  if (safe_stat (program, &st) == 0
 	      && S_ISREG (st.st_mode))
 	    {
 	      if (st.st_uid == geteuid ())
@@ -1189,12 +1267,22 @@ construct_command_argv_internal (line, restp, shell, ifs)
      char *line, **restp;
      char *shell, *ifs;
 {
+#ifdef __MSDOS__
+  static char sh_chars[] = "\"|<>";
+  static char *sh_cmds[] = { "break", "call", "cd", "chcp", "chdir", "cls",
+			     "copy", "ctty", "date", "del", "dir", "echo",
+			     "erase", "exit", "for", "goto", "if", "if", "md",
+			     "mkdir", "path", "pause", "prompt", "rem", "ren",
+			     "rename", "set", "shift", "time", "type",
+			     "ver", "verify", "vol", ":", 0 };
+#else
   static char sh_chars[] = "#;\"*?[]&|<>(){}$`^";
   static char *sh_cmds[] = { "cd", "eval", "exec", "exit", "login",
 			     "logout", "set", "umask", "wait", "while", "for",
 			     "case", "if", ":", ".", "break", "continue",
 			     "export", "read", "readonly", "shift", "times",
 			     "trap", "switch", 0 };
+#endif
   register int i;
   register char *p;
   register char *ap;
@@ -1241,6 +1329,7 @@ construct_command_argv_internal (line, restp, shell, ifs)
 
       if (instring)
 	{
+	string_char:
 	  /* Inside a string, just copy any char except a closing quote
 	     or a backslash-newline combination.  */
 	  if (*p == '\'')
@@ -1292,7 +1381,7 @@ construct_command_argv_internal (line, restp, shell, ifs)
 		  strcpy (p, p + 1);
 
 		if (instring)
-		  *ap++ = *p;
+		  goto string_char;
 		else
 		  {
 		    if (ap != new_argv[i])
@@ -1401,6 +1490,34 @@ construct_command_argv_internal (line, restp, shell, ifs)
       free (new_argv);
     }
 
+#ifdef __MSDOS__
+   {
+     FILE *batch;
+     dos_batch_file = 1;
+     if (dos_bname == 0)
+       {
+	 dos_bname = tempnam (".", "mk");
+	 for (i = 0; dos_bname[i] != '\0'; ++i)
+	   if (dos_bname[i] == '/')
+	     dos_bname[i] = '\\';
+	 dos_bename = (char *) xmalloc (strlen (dos_bname) + 5);
+	 strcpy (dos_bename, dos_bname);
+	 strcat (dos_bname, ".bat");
+	 strcat (dos_bename, ".err");
+       }
+     batch = fopen(bename, "w"); /* Create a file.  */
+     if (batch != NULL)
+       fclose (batch);
+     batch = fopen (dos_bname, "w");
+     fputs ("@echo off\n", batch);
+     fputs (line, batch);
+     fprintf (batch, "\nif errorlevel 1 del %s\n", dos_bename);
+     fclose (batch);
+     new_argv = (char **) xmalloc(2 * sizeof(char *));
+     new_argv[0] = strdup (bname);
+     new_argv[1] = 0;
+   }
+#else	/* Not MSDOS.  */
   {
     /* SHELL may be a multi-word command.  Construct a command line
        "SHELL -c LINE", with all special chars in LINE escaped.
@@ -1458,6 +1575,7 @@ construct_command_argv_internal (line, restp, shell, ifs)
     new_argv = construct_command_argv_internal (new_line, (char **) NULL,
 						(char *) 0, (char *) 0);
   }
+#endif	/* MSDOS.  */
 
   return new_argv;
 }
