@@ -46,6 +46,7 @@ static void log_working_directory ();
 static void print_data_base (), print_version ();
 static void decode_switches (), decode_env_switches ();
 static void define_makeflags ();
+static char *quote_as_word ();
 
 /* The structure that describes an accepted command switch.  */
 
@@ -271,13 +272,13 @@ static const struct command_switch switches[] =
     { 'w', flag, (char *) &print_directory_flag, 1, 1, 0, 0, 0,
 	"print-directory", 0,
 	"Print the current directory" },
-    { 1, flag, (char *) &inhibit_print_directory_flag, 1, 1, 0, 0, 0,
+    { 2, flag, (char *) &inhibit_print_directory_flag, 1, 1, 0, 0, 0,
 	"no-print-directory", 0,
 	"Turn off -w, even if it was turned on implicitly" },
     { 'W', string, (char *) &new_files, 0, 0, 0, 0, 0,
 	"what-if", "FILE",
 	"Consider FILE to be infinitely new" },
-    { 2, flag, (char *) &warn_undefined_variables_flag, 1, 1, 0, 0, 0,
+    { 3, flag, (char *) &warn_undefined_variables_flag, 1, 1, 0, 0, 0,
 	"warn-undefined-variables", 0,
 	"Warn when an undefined variable is referenced" },
     { '\0', }
@@ -304,14 +305,28 @@ static struct option long_option_aliases[] =
 
 #define	DESCRIPTION_COLUMN	30
 
-/* List of non-switch arguments.  */
+/* List of goal targets.  */
 
-struct stringlist *other_args = 0;
+static struct dep *goals, *lastgoal;
+
+/* List of variables which were defined on the command line
+   (or, equivalently, in MAKEFLAGS).  */
+
+struct command_variable
+  {
+    struct command_variable *next;
+    struct variable *variable;
+  };
+static struct command_variable *command_variables;
 
 /* The name we were invoked with.  */
 
 char *program;
 
+/* Our current directory before processing any -C options.  */
+
+char *directory_before_chdir;
+
 /* Our current directory after processing all -C options.  */
 
 char *starting_directory;
@@ -331,6 +346,11 @@ struct file *default_goal_file;
    This is zero if the makefiles do not define .DEFAULT.  */
 
 struct file *default_file;
+
+/* Nonzero if we have seen the magic `.POSIX' target.
+   This turns on pedantic compliance with POSIX.2.  */
+
+int posix_pedantic;
 
 /* Mask of signals that are being caught with fatal_error_signal.  */
 
@@ -387,14 +407,9 @@ main (argc, argv, envp)
   extern RETSIGTYPE fatal_error_signal (), child_handler ();
   register struct file *f;
   register unsigned int i;
-  register char *cmd_defs;
-  register unsigned int cmd_defs_len, cmd_defs_idx;
   char **p;
-  struct dep *goals = 0;
-  register struct dep *lastgoal;
   struct dep *read_makefiles;
   PATH_VAR (current_directory);
-  char *directory_before_chdir;
 
   default_goal_file = 0;
   reading_filename = 0;
@@ -464,6 +479,12 @@ main (argc, argv, envp)
   else 
     {
       program = rindex (argv[0], '/');
+#ifdef __MSDOS__
+      if (program == 0)
+	program = rindex (argv[0], '\\');
+      if (program == 0)
+	program = rindex (argv[0], ':');
+#endif
       if (program == 0)
 	program = argv[0];
       else
@@ -490,8 +511,8 @@ main (argc, argv, envp)
 					 strlen (current_directory));
 
   /* Read in variables from the environment.  It is important that this be
-     done before `MAKE' and `MAKEOVERRIDES' are figured out so their
-     definitions will not be ones from the environment.  */
+     done before $(MAKE) is are figured out so its definitions will not be
+     one from the environment.  */
 
   for (i = 0; envp[i] != 0; ++i)
     {
@@ -531,76 +552,7 @@ main (argc, argv, envp)
   if (print_version_flag)
     die (0);
 
-  /* Search for command line arguments that define variables,
-     and do the definitions.  Also save up the text of these
-     arguments in CMD_DEFS so we can put them into the values
-     of $(MAKEOVERRIDES) and $(MAKE).  */
-
-  cmd_defs_len = 200;
-  cmd_defs = (char *) xmalloc (cmd_defs_len);
-  cmd_defs_idx = 0;
-
-  for (i = 1; other_args->list[i] != 0; ++i)
-    {
-      if (other_args->list[i][0] == '\0')
-	/* Ignore empty arguments, so they can't kill enter_file.  */
-	continue;
-
-      /* Try a variable definition.  */
-      if (try_variable_definition ((char *) 0, 0,
-				   other_args->list[i], o_command))
-	{
-	  /* It succeeded.  The variable is already defined.
-	     Backslash-quotify it and append it to CMD_DEFS, then clobber it
-	     to 0 in the list so that it won't be taken for a goal target.  */
-	  register char *p = other_args->list[i];
-	  unsigned int l = strlen (p);
-	  if (cmd_defs_idx + (l * 2) + 1 > cmd_defs_len)
-	    {
-	      if (l * 2 > cmd_defs_len)
-		cmd_defs_len += l * 2;
-	      else
-		cmd_defs_len *= 2;
-	      cmd_defs = (char *) xrealloc (cmd_defs, cmd_defs_len);
-	    }
-	  
-	  while (*p != '\0')
-	    {
-	      if (index ("^;'\"*?[]$<>(){}|&~`\\ \t\r\n\f\v", *p) != 0)
-		cmd_defs[cmd_defs_idx++] = '\\';
-	      cmd_defs[cmd_defs_idx++] = *p++;
-	    }
-	  cmd_defs[cmd_defs_idx++] = ' ';
-	}
-      else
-	{
-	  /* It was not a variable definition, so it is a target to be made.
-	     Enter it as a file and add it to the dep chain of goals.  */
-	  f = enter_command_line_file (other_args->list[i]);
-	  f->cmd_target = 1;
-	  
-	  if (goals == 0)
-	    {
-	      goals = (struct dep *) xmalloc (sizeof (struct dep));
-	      lastgoal = goals;
-	    }
-	  else
-	    {
-	      lastgoal->next = (struct dep *) xmalloc (sizeof (struct dep));
-	      lastgoal = lastgoal->next;
-	    }
-	  lastgoal->name = 0;
-	  lastgoal->file = f;
-	}
-    }
-
-  if (cmd_defs_idx > 0)
-    {
-      cmd_defs[cmd_defs_idx - 1] = '\0';
-      (void) define_variable ("MAKEOVERRIDES", 13, cmd_defs, o_default, 0);
-    }
-  free (cmd_defs);
-
+#ifndef __MSDOS__
   /* Set the "MAKE_COMMAND" variable to the name we were invoked with.
      (If it is a relative pathname with a slash, prepend our directory name
      so the result will run the same program regardless of the current dir.
@@ -610,17 +562,62 @@ main (argc, argv, envp)
   if (current_directory[0] != '\0'
       && argv[0] != 0 && argv[0][0] != '/' && index (argv[0], '/') != 0)
     argv[0] = concat (current_directory, "/", argv[0]);
+#endif
 
+  /* The extra indirection through $(MAKE_COMMAND) is done
+     for hysterical raisins.  */
   (void) define_variable ("MAKE_COMMAND", 12, argv[0], o_default, 0);
+  (void) define_variable ("MAKE", 4, "$(MAKE_COMMAND)", o_default, 1);
 
-  /* Append the command-line variable definitions gathered above
-     so sub-makes will get them as command-line definitions.  */
+  if (command_variables != 0)
+    {
+      struct command_variable *cv;
+      struct variable *v;
+      unsigned int len = 0;
+      char *value, *p;
 
-  (void) define_variable ("MAKE", 4,
-			  "$(MAKE_COMMAND) $(MAKEOVERRIDES)", o_default, 1);
+      /* Figure out how much space will be taken up by the command-line
+	 variable definitions.  */
+      for (cv = command_variables; cv != 0; cv = cv->next)
+	{
+	  v = cv->variable;
+	  len += 2 * strlen (v->name);
+	  if (! v->recursive)
+	    ++len;
+	  ++len;
+	  len += 2 * strlen (v->value);
+	}
+
+      /* Now allocate a buffer big enough and fill it.  */
+      p = value = alloca (len);
+      for (cv = command_variables; cv != 0; cv = cv->next)
+	{
+	  v = cv->variable;
+	  p = quote_as_word (p, v->name, 0);
+	  if (! v->recursive)
+	    *p++ = ':';
+	  *p++ = '=';
+	  p = quote_as_word (p, v->value, 0);
+	  *p++ = ' ';
+	}
+      p[-1] = '\0';		/* Kill the final space and terminate.  */
+
+      /* Define an unchangeable variable with a name that no POSIX.2
+	 makefile could validly use for its own variable.  */
+      (void) define_variable ("-*-command-variables-*-", 23,
+			      value, o_automatic, 0);
+
+      /* Define the variable; this will not override any user definition.
+         Normally a reference to this variable is written into the value of
+         MAKEFLAGS, allowing the user to override this value to affect the
+         exported value of MAKEFLAGS.  In POSIX-pedantic mode, we cannot
+         allow the user's setting of MAKEOVERRIDES to affect MAKEFLAGS, so
+         a reference to this hidden variable is written instead. */
+      (void) define_variable ("MAKEOVERRIDES", 13,
+			      "${-*-command-variables-*-}", o_env, 1);
+    }
 
   /* If there were -C flags, move ourselves about.  */
-
   if (directories != 0)
     for (i = 0; directories->list[i] != 0; ++i)
       {
@@ -788,12 +785,8 @@ main (argc, argv, envp)
 
   define_makeflags (1, 0);
 
-  ignore_errors_flag |= lookup_file (".IGNORE") != 0;
-
-  silent_flag |= lookup_file (".SILENT") != 0;
-
-  /* Make each `struct dep' point at the
-     `struct file' for the file depended on.  */
+  /* Make each `struct dep' point at the `struct file' for the file
+     depended on.  Also do magic for special targets.  */
 
   snap_deps ();
 
@@ -1123,7 +1116,7 @@ main (argc, argv, envp)
 
 /* Parsing of arguments, decoding of switches.  */
 
-static char options[sizeof (switches) / sizeof (switches[0]) * 3];
+static char options[1 + sizeof (switches) / sizeof (switches[0]) * 3];
 static struct option long_options[(sizeof (switches) / sizeof (switches[0])) +
 				  (sizeof (long_option_aliases) /
 				   sizeof (long_option_aliases[0]))];
@@ -1141,6 +1134,11 @@ init_switches ()
     return;
 
   p = options;
+
+  /* Return switch and non-switch args in order, regardless of
+     POSIXLY_CORRECT.  Non-switch args are returned as option 1.  */
+  *p++ = '-';
+
   for (i = 0; switches[i].c != '\0'; ++i)
     {
       long_options[i].name = (switches[i].long_name == 0 ? "" :
@@ -1195,15 +1193,6 @@ decode_switches (argc, argv, env)
   register struct stringlist *sl;
   register int c;
 
-  if (!env)
-    {
-      other_args = (struct stringlist *) xmalloc (sizeof (struct stringlist));
-      other_args->max = argc + 1;
-      other_args->list = (char **) xmalloc ((argc + 1) * sizeof (char *));
-      other_args->idx = 1;
-      other_args->list[0] = argv[0];
-    }
-
   /* getopt does most of the parsing for us.
      First, get its vectors set up.  */
 
@@ -1215,10 +1204,65 @@ decode_switches (argc, argv, env)
   /* Reset getopt's state.  */
   optind = 0;
 
-  while ((c = getopt_long (argc, argv,
-			   options, long_options, (int *) 0)) != EOF)
+  c = 0;
+  while (optind < argc)
     {
-      if (c == '?')
+      if (c == EOF)
+	{
+	  /* There are no more options according to getting getopt, but
+	     there are some arguments left.  Since we have asked for
+	     non-option arguments to be returned in order, I think this
+	     only happens when there is a "--" argument to prevent later
+	     argument from being options.  Since getopt has finished its
+	     job, just update its state variables for the next argument and
+	     set C as if it had returned 1, indicating a non-option
+	     argument.  */
+	  optarg = argv[optind++];
+	  c = 1;
+	}
+      else
+	/* Parse the next argument.  */
+	c = getopt_long (argc, argv, options, long_options, (int *) 0);
+
+      if (c == 1)
+	{
+	  /* Non-option argument.  It might be a variable definition.  */
+	  struct variable *v;
+	  v = try_variable_definition ((char *) 0, 0, optarg, o_command);
+	  if (v != 0)
+	    {
+	      /* It is indeed a variable definition.  Record a pointer to
+		 the variable for later use in define_makeflags.  */
+	      struct command_variable *cv
+		= (struct command_variable *) xmalloc (sizeof (*cv));
+	      cv->variable = v;
+	      cv->next = command_variables;
+	      command_variables = cv;
+	    }
+	  else if (! env)
+	    {
+	      /* Not an option or variable definition; it must be a goal
+		 target!  Enter it as a file and add it to the dep chain of
+		 goals.  */
+	      struct file *f = enter_command_line_file (optarg);
+	      f->cmd_target = 1;
+	  
+	      if (goals == 0)
+		{
+		  goals = (struct dep *) xmalloc (sizeof (struct dep));
+		  lastgoal = goals;
+		}
+	      else
+		{
+		  lastgoal->next
+		    = (struct dep *) xmalloc (sizeof (struct dep));
+		  lastgoal = lastgoal->next;
+		}
+	      lastgoal->name = 0;
+	      lastgoal->file = f;
+	    }
+	}
+      else if (c == '?')
 	/* Bad option.  We will print a usage message and die later.
 	   But continue to parse the other options so the user can
 	   see all he did wrong.  */
@@ -1320,19 +1364,6 @@ positive integral argument",
 	    }
     }
 
-  if (!env)
-    {
-      /* Collect the remaining args in the `other_args' string list.  */
-
-      while (optind < argc)
-	{
-	  char *arg = argv[optind++];
-	  if (arg[0] != '-' || arg[1] != '\0')
-	    other_args->list[other_args->idx++] = arg;
-	}
-      other_args->list[other_args->idx] = 0;
-    }
-
   if (!env && (bad || print_usage_flag))
     {
       /* Print a nice usage message.  */
@@ -1437,7 +1468,7 @@ decode_env_switches (envar, len)
      unsigned int len;
 {
   char *varref = (char *) alloca (2 + len + 2);
-  char *value, *args;
+  char *value, *p;
   int argc;
   char **argv;
 
@@ -1455,37 +1486,95 @@ decode_env_switches (envar, len)
   if (len == 0)
     return;
 
-  /* Make a copy of the value in ARGS, where we will munge it.
-     If it does not begin with a dash, prepend one.
-     We must allocate lasting storage for this (and we never free it) because
-     decode_switches may save pointers into it for string-valued switches.  */
-  args = (char *) xmalloc (1 + len + 2);
-  if (value[0] != '-')
-    args[0] = '-';
-  bcopy (value, value[0] == '-' ? args : &args[1], len + 1);
-  /* Write an extra null terminator so our loop below will
-     never be in danger of looking past the end of the string.  */
-  args[(value[0] == '-' ? 0 : 1) + len + 1] = '\0';
-
   /* Allocate a vector that is definitely big enough.  */
   argv = (char **) alloca ((1 + len + 1) * sizeof (char *));
 
+  /* Allocate a buffer to copy the value into while we split it into words
+     and unquote it.  We must use permanent storage for this because
+     decode_switches may store pointers into the passed argument words.  */
+  p = (char *) xmalloc (2 * len);
+
   /* getopt will look at the arguments starting at ARGV[1].
      Prepend a spacer word.  */
   argv[0] = 0;
   argc = 1;
-  do
+  argv[argc] = p;
+  while (*value != '\0')
     {
-      argv[argc++] = args;
-      args = end_of_token (args);
-      *args++ = '\0';
-    } while (*args != '\0');
-  argv[argc] = 0;
+      if (*value == '\\')
+	++value;		/* Skip the backslash.  */
+      else if (isblank (*value))
+	{
+	  /* End of the word.  */
+	  *p++ = '\0';
+	  argv[++argc] = p;
+	  do
+	    ++value;
+	  while (isblank (*value));
+	  continue;
+	}
+      *p++ = *value++;
+    }
+  *p = '\0';
+  argv[++argc] = 0;
+
+  if (argc == 2 && argv[1][0] != '-')
+    {
+      /* There is just one word in the value, and it is not a switch.
+	 Either this is the single-word form and we should prepend a dash
+	 before calling decode_switches, or this is the multi-word form and
+	 there is no dash because it is a variable definition.  */
+      struct variable *v;
+      v = try_variable_definition ((char *) 0, 0, argv[1], o_command);
+      if (v != 0)
+	{
+	  /* It was indeed a variable definition, and now it has been
+	     processed.  There is nothing for decode_switches to do.
+	     Record a pointer to the variable for later use in
+	     define_makeflags.  */
+	  struct command_variable *cv
+	    = (struct command_variable *) xmalloc (sizeof (*cv));
+	  cv->variable = v;
+	  cv->next = command_variables;
+	  command_variables = cv;
+	  return;
+	}
+
+      /* It wasn't a variable definition, so it's some switches without a
+	 leading dash.  Add one and pass it along to decode_switches.  We
+	 need permanent storage for this in case decode_switches saves
+	 pointers into the value.  */
+      argv[1] = concat ("-", argv[1], "");
+    }
 
   /* Parse those words.  */
   decode_switches (argc, argv, 1);
 }
 
+/* Quote the string IN so that it will be interpreted as a single word with
+   no magic by the shell; if DOUBLE_DOLLARS is nonzero, also double dollar
+   signs to avoid variable expansion in make itself.  Write the result into
+   OUT, returning the address of the next character to be written.
+   Allocating space for OUT twice the length of IN (thrice if
+   DOUBLE_DOLLARS is nonzero) is always sufficient.  */
+
+static char *
+quote_as_word (out, in, double_dollars)
+     char *out, *in;
+     int double_dollars;
+{
+  while (*in != '\0')
+    {
+      if (index ("^;'\"*?[]$<>(){}|&~`\\ \t\r\n\f\v", *in) != 0)
+	*out++ = '\\';
+      if (double_dollars && *in == '$')
+	*out++ = '$';
+      *out++ = *in++;
+    }
+
+  return out;
+}
+
 /* Define the MAKEFLAGS and MFLAGS variables to reflect the settings of the
    command switches.  Include options with args if ALL is nonzero.
    Don't include options with the `no_makefile' flag set if MAKEFILE.  */
@@ -1494,8 +1583,12 @@ static void
 define_makeflags (all, makefile)
      int all, makefile;
 {
+  static const char ref[] = "$(MAKEOVERRIDES)";
+  static const char posixref[] = "$(-*-command-variables-*-)";
   register const struct command_switch *cs;
   char *flagstring;
+  register char *p;
+  unsigned int words;
   struct variable *v;
 
   /* We will construct a linked list of `struct flag's describing
@@ -1508,7 +1601,6 @@ define_makeflags (all, makefile)
       struct flag *next;
       const struct command_switch *cs;
       char *arg;
-      unsigned int arglen;
     };
   struct flag *flags = 0;
   unsigned int flagslen = 0;
@@ -1517,13 +1609,12 @@ define_makeflags (all, makefile)
     struct flag *new = (struct flag *) alloca (sizeof (struct flag));	      \
     new->cs = cs;							      \
     new->arg = (ARG);							      \
-    new->arglen = (LEN);						      \
     new->next = flags;							      \
     flags = new;							      \
     if (new->arg == 0)							      \
       ++flagslen;		/* Just a single flag letter.  */	      \
     else								      \
-      flagslen += 1 + 1 + 1 + 1 + new->arglen; /* " -x foo" */		      \
+      flagslen += 1 + 1 + 1 + 1 + 3 * (LEN); /* " -x foo" */		      \
     if (!isalnum (cs->c))						      \
       /* This switch has no single-letter version, so we use the long.  */    \
       flagslen += 2 + strlen (cs->long_name);				      \
@@ -1607,79 +1698,111 @@ define_makeflags (all, makefile)
 	  break;
 	}
 
+  flagslen += 4 + sizeof posixref; /* Four more for the possible " -- ".  */
+
 #undef	ADD_FLAG
 
-  if (flags == 0)
-    /* No flags.  Use a string of two nulls so [1] works below.  */
-    flagstring = "\0";
-  else
+  /* Construct the value in FLAGSTRING.
+     We allocate enough space for a preceding dash and trailing null.  */
+  flagstring = (char *) alloca (1 + flagslen + 1);
+  p = flagstring;
+  words = 1;
+  *p++ = '-';
+  while (flags != 0)
     {
-      /* Construct the value in FLAGSTRING.
-	 We allocate enough space for a preceding dash and trailing null.  */
-      register char *p;
-      flagstring = (char *) alloca (1 + flagslen + 1);
-      p = flagstring;
-      *p++ = '-';
-      do
+      /* Add the flag letter or name to the string.  */
+      if (!isalnum (flags->cs->c))
 	{
-	  /* Add the flag letter or name to the string.  */
-	  if (!isalnum (flags->cs->c))
+	  *p++ = '-';
+	  strcpy (p, flags->cs->long_name);
+	  p += strlen (p);
+	}
+      else
+	*p++ = flags->cs->c;
+      if (flags->arg != 0)
+	{
+	  /* A flag that takes an optional argument which in this case is
+	     omitted is specified by ARG being "".  We must distinguish
+	     because a following flag appended without an intervening " -"
+	     is considered the arg for the first.  */
+	  if (flags->arg[0] != '\0')
 	    {
-	      *p++ = '-';
-	      strcpy (p, flags->cs->long_name);
-	      p += strlen (p);
+	      /* Add its argument too.  */
+	      *p++ = !isalnum (flags->cs->c) ? '=' : ' ';
+	      p = quote_as_word (p, flags->arg, 1);
 	    }
-	  else
-	    *p++ = flags->cs->c;
-	  if (flags->arg != 0)
-	    {
-	      /* A flag that takes an optional argument which in this case
-		 is omitted is specified by ARG being "" and ARGLEN being 0.
-		 We must distinguish because a following flag appended without
-		 an intervening " -" is considered the arg for the first.  */
-	      if (flags->arglen > 0)
-		{
-		  /* Add its argument too.  */
-		  *p++ = !isalnum (flags->cs->c) ? '=' : ' ';
-		  bcopy (flags->arg, p, flags->arglen);
-		  p += flags->arglen;
-		}
-	      /* Write a following space and dash, for the next flag.  */
-	      *p++ = ' ';
-	      *p++ = '-';
-	    }
-	  else if (!isalnum (flags->cs->c))
-	    {
-	      /* Long options must each go in their own word,
-		 so we write the following space and dash.  */
-	      *p++ = ' ';
-	      *p++ = '-';
-	    }
-	  flags = flags->next;
-	} while (flags != 0);
-
-      if (p[-1] == '-')
-	/* Kill the final space and dash.  */
-	p -= 2;
-
-      /* Terminate the string.  */
-      *p = '\0';
+	  ++words;
+	  /* Write a following space and dash, for the next flag.  */
+	  *p++ = ' ';
+	  *p++ = '-';
+	}
+      else if (!isalnum (flags->cs->c))
+	{
+	  ++words;
+	  /* Long options must each go in their own word,
+	     so we write the following space and dash.  */
+	  *p++ = ' ';
+	  *p++ = '-';
+	}
+      flags = flags->next;
     }
 
+  if (all && command_variables != 0)
+    {
+      /* Now write a reference to $(MAKEOVERRIDES), which contains all the
+	 command-line variable definitions.  */
+
+      if (p == &flagstring[1])
+	/* No flags written, so elide the leading dash already written.  */
+	p = flagstring;
+      else
+	{
+	  /* Separate the variables from the switches with a "--" arg.  */
+	  if (p[-1] != '-')
+	    {
+	      /* We did not already write a trailing " -".  */
+	      *p++ = ' ';
+	      *p++ = '-';
+	    }
+	  /* There is a trailing " -"; fill it out to " -- ".  */
+	  *p++ = '-';
+	  *p++ = ' ';
+	}
+
+      /* Copy in the string.  */
+      if (posix_pedantic)
+	{
+	  bcopy (posixref, p, sizeof posixref - 1);
+	  p += sizeof posixref - 1;
+	}
+      else
+	{
+	  bcopy (ref, p, sizeof ref - 1);
+	  p += sizeof ref - 1;
+	}
+    }
+  else if (p == &flagstring[1])
+    --p;
+  else if (p[-1] == '-')
+    /* Kill the final space and dash.  */
+    p -= 2;
+  /* Terminate the string.  */
+  *p = '\0';
+
   v = define_variable ("MAKEFLAGS", 9,
-		       /* On Sun, the value of MFLAGS starts with a `-' but
-			  the value of MAKEFLAGS lacks the `-'.
-			  Be compatible with this unless FLAGSTRING starts
-			  with a long option `--foo', since removing the
-			  first dash would result in the bogus `-foo'.  */
-		       flagstring[1] == '-' ? flagstring : &flagstring[1],
+		       /* If there is just a single word of switches,
+			  omit the leading dash unless it is a single
+			  long option with two leading dashes.  */
+		       &flagstring[(words == 1 && command_variables == 0
+				    && flagstring[1] != '-')
+				   ? 1 : 0],
 		       /* This used to use o_env, but that lost when a
 			  makefile defined MAKEFLAGS.  Makefiles set
 			  MAKEFLAGS to add switches, but we still want
 			  to redefine its value with the full set of
 			  switches.  Of course, an override or command
 			  definition will still take precedence.  */
-		       o_file, 0);
+		       o_file, 1);
   if (! all)
     /* The first time we are called, set MAKEFLAGS to always be exported.
        We should not do this again on the second call, because that is
@@ -1687,7 +1810,7 @@ define_makeflags (all, makefile)
     v->export = v_export;
   /* Since MFLAGS is not parsed for flags, there is no reason to
      override any makefile redefinition.  */
-  (void) define_variable ("MFLAGS", 6, flagstring, o_env, 0);
+  (void) define_variable ("MFLAGS", 6, flagstring, o_env, 1);
 }
 
 /* Print version information.  */
@@ -1755,6 +1878,13 @@ die (status)
 
       dying = 1;
 
+      /* Try to move back to the original directory.  This is essential on
+	 MS-DOS (where there is really only one process), and on Unix it
+	 puts core files in the original directory instead of the -C
+	 directory.  */
+      if (directory_before_chdir != 0)
+	chdir (directory_before_chdir);
+
       if (print_version_flag)
 	print_version ();
 
