@@ -1,5 +1,5 @@
-/* Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993
-	Free Software Foundation, Inc.
+/* Miscellaneous global declarations and portability cruft for GNU Make.
+Copyright (C) 1988, 89, 90, 91, 92, 93, 94 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -59,13 +59,15 @@ extern int errno;
 
 #ifdef	HAVE_UNISTD_H
 #include <unistd.h>
-#ifdef	_POSIX_VERSION
+/* Ultrix's unistd.h always defines _POSIX_VERSION, but you only get
+   POSIX.1 behavior with `cc -YPOSIX', which predefines POSIX itself!  */
+#if defined (_POSIX_VERSION) && !defined (ultrix)
 #define	POSIX
 #endif
 #endif
 
 /* Some systems define _POSIX_VERSION but are not really POSIX.1.  */
-#if (defined (butterfly) || \
+#if (defined (butterfly) || defined (__arm) || \
      (defined (__mips) && defined (_SYSTYPE_SVR3)) || \
      (defined (sequent) && defined (i386)))
 #undef POSIX
@@ -210,6 +212,13 @@ extern void bcopy ();
 #endif	/* ANSI_STRING.  */
 #undef	ANSI_STRING
 
+/* SCO Xenix has a buggy macro definition in <string.h>.  */
+#undef	strerror
+
+#ifndef ANSI_STRING
+extern char *strerror ();
+#endif
+
 
 #ifdef	__GNUC__
 #undef	alloca
@@ -243,6 +252,12 @@ extern char *alloca ();
 #define	ENUM_BITFIELD(bits)
 #endif
 
+#ifdef __MSDOS__
+#define PATH_SEPARATOR_CHAR ';'
+#else
+#define PATH_SEPARATOR_CHAR ':'
+#endif
+
 extern void die ();
 extern void message (), fatal (), error ();
 extern void makefile_error (), makefile_fatal ();
@@ -255,7 +270,7 @@ extern char *sindex (), *lindex ();
 extern int alpha_compare ();
 extern void print_spaces ();
 extern struct dep *copy_dep_chain ();
-extern char *find_percent ();
+extern char *find_char_unquote (), *find_percent ();
 
 #ifndef	NO_ARCHIVES
 extern int ar_name ();
@@ -290,17 +305,12 @@ extern void user_access (), make_access (), child_access ();
 #include <vfork.h>
 #endif
 
-#if !defined (__GNU_LIBRARY__) && !defined (POSIX)
+/* We omit these declarations on non-POSIX systems which define _POSIX_VERSION,
+   because such systems often declare the in header files anyway.  */
+
+#if !defined (__GNU_LIBRARY__) && !defined (POSIX) && !defined (_POSIX_VERSION)
 
-#ifdef	HAVE_SIGSETMASK
-extern int sigsetmask ();
-extern int sigblock ();
-#endif
-extern int kill ();
-extern int atoi ();
 extern long int atol ();
-extern int unlink (), stat (), fstat ();
-extern int pipe (), close (), read (), write (), open ();
 extern long int lseek ();
 
 #endif	/* Not GNU C library or POSIX.  */
@@ -321,6 +331,7 @@ extern int just_print_flag, silent_flag, ignore_errors_flag, keep_going_flag;
 extern int debug_flag, print_data_base_flag, question_flag, touch_flag;
 extern int env_overrides, no_builtin_rules_flag, print_version_flag;
 extern int print_directory_flag, warn_undefined_variables_flag;
+extern int posix_pedantic;
 
 extern unsigned int job_slots;
 extern double max_load_average;
