@@ -1,4 +1,5 @@
-/* Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993 Free Software Foundation, Inc.
+/* Miscellaneous generic support functions for GNU Make.
+Copyright (C) 1988, 89, 90, 91, 92, 93, 94 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -123,36 +124,13 @@ void
 remove_comments (line)
      char *line;
 {
-  register char *p, *p2;
-  register int backslash;
-  register unsigned int bs_write;
+  char *comment;
 
-  while (1)
-    {
-      p = index (line, '#');
-      if (p == 0)
-	break;
+  comment = find_char_unquote (line, '#', 0);
 
-      backslash = 0;
-      bs_write = 0;
-      for (p2 = p - 1; p2 >= line && *p2 == '\\'; --p2)
-	{
-	  if (backslash)
-	    ++bs_write;
-	  backslash = !backslash;
-	}
-
-      if (!backslash)
-	{
-	  /* Cut off the line at the #.  */
-	  *p = '\0';
-	  break;
-	}
-
-      /* strcpy better copy left to right.  */
-      line = p;
-      strcpy (p2 + 1 + bs_write, line);
-    }
+  if (comment != 0)
+    /* Cut off the line at the #.  */
+    *comment = '\0';
 }
 
 /* Print N spaces (used by DEBUGPR for target-depth).  */
@@ -222,7 +200,7 @@ fatal (s1, s2, s3, s4, s5, s6)
   fprintf (stderr, s1, s2, s3, s4, s5, s6);
   fputs (".  Stop.\n", stderr);
 
-  die (1);
+  die (2);
 }
 
 /* Print error message.  `s1' is printf control string, `s2' is arg for it. */
@@ -264,10 +242,13 @@ makefile_fatal (file, lineno, s1, s2, s3, s4, s5, s6)
   fprintf (stderr, s1, s2, s3, s4, s5, s6);
   fputs (".  Stop.\n", stderr);
 
-  die (1);
+  die (2);
 }
 
 #ifndef HAVE_STRERROR
+
+#undef	strerror
+
 char *
 strerror (errnum)
      int errnum;
@@ -392,25 +373,9 @@ char *
 end_of_token (s)
      char *s;
 {
-  register char *p = s;
-  register int backslash = 0;
-
-  while (*p != '\0' && (backslash || !isblank (*p)))
-    {
-      if (*p++ == '\\')
-	{
-	  backslash = !backslash;
-	  while (*p == '\\')
-	    {
-	      backslash = !backslash;
-	      ++p;
-	    }
-	}
-      else
-	backslash = 0;
-    }
-
-  return p;
+  while (*s != '\0' && !isblank (*s))
+    ++s;
+  return s;
 }
 
 /* Return the address of the first nonwhitespace or null in the string S.  */
@@ -488,15 +453,46 @@ dep_name (dep)
 
 #ifdef	GETLOADAVG_PRIVILEGED
 
+#ifdef POSIX
+
+/* Hopefully if a system says it's POSIX.1 and has the setuid and setgid
+   functions, they work as POSIX.1 says.  Some systems (Alpha OSF/1 1.2,
+   for example) which claim to be POSIX.1 also have the BSD setreuid and
+   setregid functions, but they don't work as in BSD and only the POSIX.1
+   way works.  */
+
+#undef HAVE_SETREUID
+#undef HAVE_SETREGID
+
+#else	/* Not POSIX.  */
+
+/* Some POSIX.1 systems have the seteuid and setegid functions.  In a
+   POSIX-like system, they are the best thing to use.  However, some
+   non-POSIX systems have them too but they do not work in the POSIX style
+   and we must use setreuid and setregid instead.  */
+
+#undef HAVE_SETEUID
+#undef HAVE_SETEGID
+
+#endif	/* POSIX.  */
+
 #ifndef	HAVE_UNISTD_H
 extern int getuid (), getgid (), geteuid (), getegid ();
 extern int setuid (), setgid ();
+#ifdef HAVE_SETEUID
+extern int seteuid ();
+#else
 #ifdef	HAVE_SETREUID
 extern int setreuid ();
 #endif	/* Have setreuid.  */
+#endif	/* Have seteuid.  */
+#ifdef HAVE_SETEGID
+extern int setegid ();
+#else
 #ifdef	HAVE_SETREGID
 extern int setregid ();
 #endif	/* Have setregid.  */
+#endif	/* Have setegid.  */
 #endif	/* No <unistd.h>.  */
 
 /* Keep track of the user and group IDs for user- and make- access.  */
@@ -562,6 +558,16 @@ user_access ()
      We now want to set the effective user and group IDs to the real IDs,
      which are the IDs of the process that exec'd make.  */
 
+#ifdef	HAVE_SETEUID
+
+  /* Modern systems have the seteuid/setegid calls which set only the
+     effective IDs, which is ideal.  */
+
+  if (seteuid (user_uid) < 0)
+    pfatal_with_name ("user_access: seteuid");
+
+#else	/* Not HAVE_SETEUID.  */
+
 #ifndef	HAVE_SETREUID
 
   /* System V has only the setuid/setgid calls to set user/group IDs.
@@ -575,7 +581,7 @@ user_access ()
   if (setuid (user_uid) < 0)
     pfatal_with_name ("user_access: setuid");
 
-#else
+#else	/* HAVE_SETREUID.  */
 
   /* In 4BSD, the setreuid/setregid calls set both the real and effective IDs.
      They may be set to themselves or each other.  So you have two alternatives
@@ -587,14 +593,20 @@ user_access ()
   if (setreuid (make_uid, user_uid) < 0)
     pfatal_with_name ("user_access: setreuid");
 
-#endif
+#endif	/* Not HAVE_SETREUID.  */
+#endif	/* HAVE_SETEUID.  */
 
+#ifdef	HAVE_SETEGID
+  if (setegid (user_gid) < 0)
+    pfatal_with_name ("user_access: setegid");
+#else
 #ifndef	HAVE_SETREGID
   if (setgid (user_gid) < 0)
     pfatal_with_name ("user_access: setgid");
 #else
   if (setregid (make_gid, user_gid) < 0)
     pfatal_with_name ("user_access: setregid");
+#endif
 #endif
 
   current_access = user;
@@ -619,6 +631,10 @@ make_access ()
 
   /* See comments in user_access, above.  */
 
+#ifdef	HAVE_SETEUID
+  if (seteuid (make_uid) < 0)
+    pfatal_with_name ("make_access: seteuid");
+#else
 #ifndef	HAVE_SETREUID
   if (setuid (make_uid) < 0)
     pfatal_with_name ("make_access: setuid");
@@ -626,13 +642,19 @@ make_access ()
   if (setreuid (user_uid, make_uid) < 0)
     pfatal_with_name ("make_access: setreuid");
 #endif
+#endif
 
+#ifdef	HAVE_SETEGID
+  if (setegid (make_gid) < 0)
+    pfatal_with_name ("make_access: setegid");
+#else
 #ifndef	HAVE_SETREGID
   if (setgid (make_gid) < 0)
     pfatal_with_name ("make_access: setgid");
 #else
   if (setregid (user_gid, make_gid) < 0)
     pfatal_with_name ("make_access: setregid");
+#endif
 #endif
 
   current_access = make;
@@ -694,3 +716,23 @@ get_path_max ()
   return value;
 }
 #endif
+
+/* On some systems, stat can return EINTR.  */
+
+int
+safe_stat (name, buf)
+     char *name;
+     struct stat *buf;
+{
+  int ret;
+
+#ifdef EINTR
+  do
+#endif
+    ret = stat (name, buf);
+#ifdef EINTR
+  while (ret < 0 && errno == EINTR);
+#endif
+
+  return  ret;
+}
