@@ -1,5 +1,5 @@
-/* Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993
-	Free Software Foundation, Inc.
+/* Reading and parsing of makefiles for GNU Make.
+Copyright (C) 1988, 89, 90, 91, 92, 93, 94 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -33,7 +33,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 struct passwd *getpwnam ();
 
 
-static void read_makefile ();
+static int read_makefile ();
 static unsigned int readline (), do_define ();
 static int conditional_line ();
 static void record_files ();
@@ -143,7 +143,8 @@ read_all_makefiles (makefiles)
       {
 	if (*p != '\0')
 	  *p++ = '\0';
-	read_makefile (name, RM_NO_DEFAULT_GOAL | RM_INCLUDED | RM_DONTCARE);
+	(void) read_makefile (name,
+			      RM_NO_DEFAULT_GOAL | RM_INCLUDED | RM_DONTCARE);
       }
 
     free (value);
@@ -157,7 +158,8 @@ read_all_makefiles (makefiles)
 	struct dep *tail = read_makefiles;
 	register struct dep *d;
 
-	read_makefile (*makefiles, 0);
+	if (! read_makefile (*makefiles, 0))
+	  perror_with_name ("", *makefiles);
 
 	/* Find the right element of read_makefiles.  */
 	d = read_makefiles;
@@ -181,7 +183,10 @@ read_all_makefiles (makefiles)
 	++p;
 
       if (*p != 0)
-	read_makefile (*p, 0);
+	{
+	  if (! read_makefile (*p, 0))
+	    perror_with_name ("", *p);
+	}
       else
 	{
 	  /* No default makefile was found.  Add the default makefiles to the
@@ -214,9 +219,11 @@ read_all_makefiles (makefiles)
 
    FLAGS contains bits as above.
 
-   FILENAME is added to the `read_makefiles' chain.  */
+   FILENAME is added to the `read_makefiles' chain.
 
-static void
+   Returns 1 if a file was found and read, 0 if not.  */
+
+static int
 read_makefile (filename, flags)
      char *filename;
      int flags;
@@ -233,6 +240,7 @@ read_makefile (filename, flags)
   char *p2;
   int ignoring = 0, in_ignored_define = 0;
   int no_targets = 0;		/* Set when reading a rule without targets.  */
+  char *passed_filename = filename;
 
   struct nameseq *filenames = 0;
   struct dep *deps;
@@ -260,6 +268,20 @@ read_makefile (filename, flags)
   two_colon = 0;
 #endif
 
+  if (debug_flag)
+    {
+      printf ("Reading makefile `%s'", filename);
+      if (flags & RM_NO_DEFAULT_GOAL)
+	printf (" (no default goal)");
+      if (flags & RM_INCLUDED)
+	printf (" (search path)");
+      if (flags & RM_DONTCARE)
+	printf (" (don't care)");
+      if (flags & RM_NO_TILDE)
+	printf (" (no ~ expansion)");
+      puts ("...");
+    }
+
   /* First, get a stream to read.  */
 
   /* Expand ~ in FILENAME unless it came from `include',
@@ -267,7 +289,6 @@ read_makefile (filename, flags)
   if (!(flags & RM_NO_TILDE) && filename[0] == '~')
     {
       char *expanded = tilde_expand (filename);
-      /* This is a possible memory leak, but I don't care.  */
       if (expanded != 0)
 	filename = expanded;
     }
@@ -309,24 +330,21 @@ read_makefile (filename, flags)
       if (flags & RM_DONTCARE)
 	deps->file->dontcare = 1;
     }
+  if (filename != passed_filename)
+    free (filename);
   filename = deps->file->name;
-  deps->file->precious = 1;
   deps->changed = flags;
   deps = 0;
 
-  /* If the makefile can't be found at all,
-     either ignore it or give up entirely.  */
+  /* If the makefile can't be found at all, give up entirely.  */
 
   if (infile == 0)
     {
-      if (! (flags & RM_DONTCARE))
-	{
-	  /* If we did some searching, errno has the error
-	     from the last attempt, rather from FILENAME itself.  */
-	  errno = makefile_errno;
-	  perror_with_name ("fopen: ", filename);
-	}
-      return;
+      /* If we did some searching, errno has the error from the last
+	 attempt, rather from FILENAME itself.  Restore it in case the
+	 caller wants to use it in a message.  */
+      errno = makefile_errno;
+      return 0;
     }
 
   reading_filename = filename;
@@ -344,6 +362,47 @@ read_makefile (filename, flags)
       lineno += nlines;
       nlines = readline (&lb, infile, filename, lineno);
 
+      /* Check for a shell command line first.
+	 If it is not one, we can stop treating tab specially.  */
+      if (lb.buffer[0] == '\t')
+	{
+	  /* This line is a probably shell command.  */
+	  unsigned int len;
+
+	  if (no_targets)
+	    /* Ignore the commands in a rule with no targets.  */
+	    continue;
+
+	  /* If there is no preceding rule line, don't treat this line
+	     as a command, even though it begins with a tab character.
+	     SunOS 4 make appears to behave this way.  */
+
+	  if (filenames != 0)
+	    {
+	      if (ignoring)
+		/* Yep, this is a shell command, and we don't care.  */
+		continue;
+
+	      /* Append this command line to the line being accumulated.  */
+	      p = lb.buffer;
+	      if (commands_idx == 0)
+		commands_started = lineno;
+	      len = strlen (p);
+	      if (len + 1 + commands_idx > commands_len)
+		{
+		  commands_len = (len + 1 + commands_idx) * 2;
+		  commands = (char *) xrealloc (commands, commands_len);
+		}
+	      bcopy (p, &commands[commands_idx], len);
+	      commands_idx += len;
+	      commands[commands_idx++] = '\n';
+
+	      continue;
+	    }
+	}
+
+      /* This line is not a shell command line.  Don't worry about tabs.  */
+
       if (collapsed_length < lb.size)
 	{
 	  collapsed_length = lb.size;
@@ -356,17 +415,20 @@ read_makefile (filename, flags)
       collapse_continuations (collapsed);
       remove_comments (collapsed);
 
-      p = collapsed;
-      while (isspace (*p))
-	++p;
-      /* We cannot consider a line containing just a tab to be empty
-	 because it might constitute an empty command for a target.  */
-      if (*p == '\0' && lb.buffer[0] != '\t')
-	continue;
-
       /* strncmp is first to avoid dereferencing out into space.  */
 #define	word1eq(s, l) 	(!strncmp (s, p, l) \
 			 && (p[l] == '\0' || isblank (p[l])))
+      p = collapsed;
+      while (isspace (*p))
+	++p;
+      if (*p == '\0')
+	/* This line is completely empty.  */
+	continue;
+
+      /* We must first check for conditional and `define' directives before
+	 ignoring anything, since they control what we will do with
+	 following lines.  */
+
       if (!in_ignored_define
 	  && (word1eq ("ifdef", 5) || word1eq ("ifndef", 6)
 	      || word1eq ("ifeq", 4) || word1eq ("ifneq", 5)
@@ -395,7 +457,13 @@ read_makefile (filename, flags)
 	  else
 	    {
 	      p2 = next_token (p + 6);
-	      p = end_of_token (p2);
+	      /* Let the variable name be the whole rest of the line,
+		 with trailing blanks stripped (comments have already been
+		 removed), so it could be a complex variable/function
+		 reference that might contain blanks.  */
+	      p = index (p2, '\0');
+	      while (isblank (p[-1]))
+		--p;
 	      lineno = do_define (p2, p - p2, o_file,
 				  lineno, infile, filename);
 	    }
@@ -412,10 +480,15 @@ read_makefile (filename, flags)
 		in_ignored_define = 1;
 	      else
 		{
-		  unsigned int len;
-		  p2 = end_of_token (p2);
-		  p = find_next_token (&p2, &len);
-		  lineno = do_define (p, len, o_override,
+		  p2 = next_token (p2 + 6);
+		  /* Let the variable name be the whole rest of the line,
+		     with trailing blanks stripped (comments have already been
+		     removed), so it could be a complex variable/function
+		     reference that might contain blanks.  */
+		  p = index (p2, '\0');
+		  while (isblank (p[-1]))
+		    --p;
+		  lineno = do_define (p2, p - p2, o_override,
 				      lineno, infile, filename);
 		}
 	    }
@@ -427,45 +500,11 @@ read_makefile (filename, flags)
 	  continue;
 	}
 
-
       if (ignoring)
 	/* Ignore the line.  We continue here so conditionals
 	   can appear in the middle of a rule.  */
 	continue;
-      else if (lb.buffer[0] == '\t')
-	{
-	  /* This line is a shell command.  */
-	  unsigned int len;
-
-	  if (no_targets)
-	    /* Ignore the commands in a rule with no targets.  */
-	    continue;
-
-	  /* If there is no preceding rule line, don't treat this line
-	     as a command, even though it begins with a tab character.
-	     SunOS 4 make appears to behave this way.  */
-
-	  if (filenames != 0)
-	    {
-	      /* Append this command line to the line being accumulated.  */
-	      p = lb.buffer;
-	      if (commands_idx == 0)
-		commands_started = lineno;
-	      len = strlen (p);
-	      if (len + 1 + commands_idx > commands_len)
-		{
-		  commands_len = (len + 1 + commands_idx) * 2;
-		  commands = (char *) xrealloc (commands, commands_len);
-		}
-	      bcopy (p, &commands[commands_idx], len);
-	      commands_idx += len;
-	      commands[commands_idx++] = '\n';
-
-	      continue;
-	    }
-	}
-
-      if (word1eq ("export", 6))
+      else if (word1eq ("export", 6))
 	{
 	  struct variable *v;
 	  p2 = next_token (p + 6);
@@ -548,10 +587,19 @@ read_makefile (filename, flags)
 	      free (files);
 	      files = next;
 
-	      read_makefile (name, (RM_INCLUDED | RM_NO_TILDE
-				    | (noerror ? RM_DONTCARE : 0)));
+	      if (! read_makefile (name, (RM_INCLUDED | RM_NO_TILDE
+					  | (noerror ? RM_DONTCARE : 0)))
+		  && ! noerror)
+		makefile_error (filename, lineno,
+				"%s: %s", name, strerror (errno));
 	    }
 
+	  /* Free any space allocated by conditional_line.  */
+	  if (conditionals->ignoring)
+	    free (conditionals->ignoring);
+	  if (conditionals->seen_else)
+	    free (conditionals->seen_else);
+
 	  /* Restore state.  */
 	  conditionals = save;
 	  reading_filename = filename;
@@ -581,6 +629,21 @@ read_makefile (filename, flags)
       else if (try_variable_definition (filename, lineno, p, o_file))
 	/* This line has been dealt with.  */
 	;
+      else if (lb.buffer[0] == '\t')
+	{
+	  p = lb.buffer;
+	  while (isblank (*p))
+	    ++p;
+	  if (*p == '\0')
+	    /* The line is completely blank; that is harmless.  */
+	    continue;
+	  /* This line starts with a tab but was not caught above
+	     because there was no preceding target, and the line
+	     might have been usable as a variable definition.
+	     But now it is definitely lossage.  */
+	  makefile_fatal (filename, lineno,
+			  "commands commence before first target");
+	}
       else
 	{
 	  /* This line describes some target files.  */
@@ -724,6 +787,8 @@ read_makefile (filename, flags)
 
   reading_filename = 0;
   reading_lineno_ptr = 0;
+
+  return 1;
 }
 
 /* Execute a `define' directive.
@@ -759,12 +824,13 @@ do_define (name, namelen, origin, lineno, infile, filename)
     {
       lineno += nlines;
       nlines = readline (&lb, infile, filename, lineno);
-      p = next_token (lb.buffer);
 
+      collapse_continuations (lb.buffer);
+
+      p = next_token (lb.buffer);
       if ((p[5] == '\0' || isblank (p[5])) && !strncmp (p, "endef", 5))
 	{
 	  p += 5;
-	  collapse_continuations (p);
 	  remove_comments (p);
 	  if (*next_token (p) != '\0')
 	    makefile_error (filename, lineno,
@@ -781,7 +847,7 @@ do_define (name, namelen, origin, lineno, infile, filename)
 	}
       else
 	{
-	  unsigned int len = strlen (p);
+	  unsigned int len = strlen (lb.buffer);
 
 	  /* Increase the buffer size if necessary.  */
 	  if (idx + len + 1 > length)
@@ -790,7 +856,7 @@ do_define (name, namelen, origin, lineno, infile, filename)
 	      definition = (char *) xrealloc (definition, length + 1);
 	    }
 
-	  bcopy (p, &definition[idx], len);
+	  bcopy (lb.buffer, &definition[idx], len);
 	  idx += len;
 	  /* Separate lines with a newline.  */
 	  definition[idx++] = '\n';
@@ -1291,11 +1357,15 @@ record_files (filenames, pattern, pattern_percent, deps, commands_started,
 			    "target file `%s' has both : and :: entries",
 			    f->name);
 	  f = enter_file (name);
-	  /* If there was an existing entry and it was a
-	     double-colon entry, enter_file will have returned a
-	     new one, making it the prev pointer of the old one.  */
+	  /* If there was an existing entry and it was a double-colon
+	     entry, enter_file will have returned a new one, making it the
+	     prev pointer of the old one, and setting its double_colon
+	     pointer to the first one.  */
+	  if (f->double_colon == 0)
+	    /* This is the first entry for this name, so we must
+	       set its double_colon pointer to itself.  */
+	    f->double_colon = f;
 	  f->is_target = 1;
-	  f->double_colon = 1;
 	  f->deps = this;
 	  f->cmds = cmds;
 	}
@@ -1355,83 +1425,80 @@ record_files (filenames, pattern, pattern_percent, deps, commands_started,
     }
 }
 
+/* Search STRING for an unquoted STOPCHAR or blank (if BLANK is nonzero).
+   Backslashes quote STOPCHAR, blanks if BLANK is nonzero, and backslash.
+   Quoting backslashes are removed from STRING by compacting it into
+   itself.  Returns a pointer to the first unquoted STOPCHAR if there is
+   one, or nil if there are none.  */
+
+char *
+find_char_unquote (string, stopchar, blank)
+     char *string;
+     int stopchar;
+     int blank;
+{
+  unsigned int string_len = strlen (string);
+  register char *p = string;
+
+  while (1)
+    {
+      if (blank)
+	{
+	  while (*p != '\0' && *p != stopchar && !isblank (*p))
+	    ++p;
+	  if (*p == '\0')
+	    break;
+	}
+      else
+	{
+	  p = index (p, stopchar);
+	  if (p == 0)
+	    break;
+	}
+      if (p > string && p[-1] == '\\')
+	{
+	  /* Search for more backslashes.  */
+	  register int i = -2;
+	  while (&p[i] >= string && p[i] == '\\')
+	    --i;
+	  ++i;
+	  /* The number of backslashes is now -I.
+	     Copy P over itself to swallow half of them.  */
+	  bcopy (&p[i / 2], &p[i], (string_len - (p - string)) - (i / 2) + 1);
+	  p += i / 2;
+	  if (i % 2 == 0)
+	    /* All the backslashes quoted each other; the STOPCHAR was
+	       unquoted.  */
+	    return p;
+
+	  /* The STOPCHAR was quoted by a backslash.  Look for another.  */
+	}
+      else
+	/* No backslash in sight.  */
+	return p;
+    }
+
+  /* Never hit a STOPCHAR or blank (with BLANK nonzero).  */
+  return 0;
+}
+
+/* Search PATTERN for an unquoted %.  */
+
+char *
+find_percent (pattern)
+     char *pattern;
+{
+  return find_char_unquote (pattern, '%', 0);
+}
+
 /* Search STRING for an unquoted ; that is not after an unquoted #.  */
 
 static char *
 find_semicolon (string)
      char *string;
 {
-  char *found, *p;
-
-  found = index (string, ';');
-  while (found != 0 && found[-1] == '\\')
-    {
-      register char *q = &found[-1];
-      register int backslash = 0;
-      while (*q-- == '\\')
-	backslash = !backslash;
-      if (backslash)
-	found = index (found + 1, ';');
-      else
-	break;
-    }
-  if (found == 0)
-    return 0;
-
-  /* Look for a comment character (#) before the ; we found.  */
-  p = lindex (string, found, '#');
-  while (p != 0 && p[-1] == '\\')
-    {
-      register char *q = &p[-1];
-      register int backslash = 0;
-      while (*q-- == '\\')
-	backslash = !backslash;
-      if (backslash)
-	p = lindex (p + 1, found, '#');
-      else
-	break;
-    }
-  if (p == 0)
-    return found;
-  return 0;
-}
-
-/* Search PATTERN for an unquoted %.  Backslashes quote % and backslash.
-   Quoting backslashes are removed from PATTERN by compacting it into
-   itself.  Returns a pointer to the first unquoted % if there is one,
-   or nil if there are none.  */
-
-char *
-find_percent (pattern)
-     char *pattern;
-{
-  unsigned int pattern_len = strlen (pattern);
-  register char *p = pattern;
-
-  while ((p = index (p, '%')) != 0)
-    if (p > pattern && p[-1] == '\\')
-      {
-	/* Search for more backslashes.  */
-	register int i = -2;
-	while (&p[i] >= pattern && p[i] == '\\')
-	  --i;
-	++i;
-	/* The number of backslashes is now -I.
-	   Copy P over itself to swallow half of them.  */
-	bcopy (&p[i / 2], &p[i], (pattern_len - (p - pattern)) - (i / 2) + 1);
-	p += i / 2;
-	if (i % 2 == 0)
-	  /* All the backslashes quoted each other; the % was unquoted.  */
-	  return p;
-
-	/* The % was quoted by a backslash.  Look for another.  */
-      }
-    else
-      /* No backslash in sight.  */
-      return p;
-
-  /* Never hit a %.  */
-  return 0;
+  return (find_char_unquote (string, '#', 0) == 0
+	  ? find_char_unquote (string, ';', 0) : 0);
 }
 
 /* Parse a string into a sequence of filenames represented as a
@@ -1455,11 +1522,10 @@ parse_file_seq (stringp, stopchar, size, strip)
      int strip;
 {
   register struct nameseq *new = 0;
-  register struct nameseq *new1;
+  register struct nameseq *new1, *lastnew1;
   register char *p = *stringp;
   char *q;
   char *name;
-  register int c;
 
   while (1)
     {
@@ -1471,18 +1537,9 @@ parse_file_seq (stringp, stopchar, size, strip)
 	break;
       /* Yes, find end of next name.  */
       q = p;
-      while (1)
-	{
-	  c = *p++;
-	  if (c == '\0')
-	    break;
-	  else if (c == '\\' &&
-	           (*p == '\\' || isblank (*p) || *p == stopchar))
-	    ++p;
-	  else if (isblank (c) || c == stopchar)
-	    break;
-	}
-      p--;
+      p = find_char_unquote (q, stopchar, 1);
+      if (p == 0)
+	p = q + strlen (q);
 
       if (strip)
 	/* Skip leading `./'s.  */
@@ -1516,7 +1573,7 @@ parse_file_seq (stringp, stopchar, size, strip)
      an elt further down the chain (i.e., previous in the file list)
      with an unmatched `(' (e.g., "lib(mem").  */
 
-  for (new1 = new; new1 != 0; new1 = new1->next)
+  for (new1 = new, lastnew1 = 0; new1 != 0; lastnew1 = new1, new1 = new1->next)
     if (new1->name[0] != '('	/* Don't catch "(%)" and suchlike.  */
 	&& new1->name[strlen (new1->name) - 1] == ')'
 	&& index (new1->name, '(') == 0)
@@ -1531,7 +1588,11 @@ parse_file_seq (stringp, stopchar, size, strip)
 	    lastn = n;
 	    n = n->next;
 	  }
-	if (n != 0)
+	if (n != 0
+	    /* Ignore something starting with `(', as that cannot actually
+	       be an archive-member reference (and treating it as such
+	       results in an empty file name, which causes much lossage).  */
+	    && n->name[0] != '(')
 	  {
 	    /* N is the first element in the archive group.
 	       Its name looks like "lib(mem" (with no closing `)').  */
@@ -1566,10 +1627,12 @@ parse_file_seq (stringp, stopchar, size, strip)
 	      {
 		/* NEW1 is just ")", part of something like "lib(a b )".
 		   Omit it from the chain and free its storage.  */
+		if (lastnew1 == 0)
+		  new = new1->next;
+		else
+		  lastnew1->next = new1->next;
 		lastn = new1;
 		new1 = new1->next;
-		if (new == lastn)
-		  new = new1;
 		free (lastn->name);
 		free ((char *) lastn);
 	      }
@@ -1593,6 +1656,11 @@ parse_file_seq (stringp, stopchar, size, strip)
 		new1->name = name;
 		new1 = new1->next;
 	      }
+
+	    if (new1 == 0)
+	      /* We might have slurped up the whole list,
+		 and continuing the loop would dereference NEW1.  */
+	      break;
 	  }
       }
 
@@ -1624,18 +1692,21 @@ readline (linebuffer, stream, filename, lineno)
 
   *p = '\0';
 
-  while (1)
+  while (fgets (p, end - p, stream) != 0)
     {
-      if (fgets (p, end - p, stream) == 0)
-	if (feof (stream))
-	  break;
-	else
-	  pfatal_with_name (filename);
-
       len = strlen (p);
       if (len == 0)
-	/* This only happens when the first thing on the line is a '\0'.  */
-	makefile_fatal (filename, lineno, "NUL not allowed in makefile");
+	{
+	  /* This only happens when the first thing on the line is a '\0'.
+	     It is a pretty hopeless case, but (wonder of wonders) Athena
+	     lossage strikes again!  (xmkmf puts NULs in its makefiles.)
+	     There is nothing really to be done; we synthesize a newline so
+	     the following line doesn't appear to be part of this line.  */
+	  makefile_error (filename, lineno,
+			  "warning: NUL character seen; rest of line ignored");
+	  p[0] = '\n';
+	  len = 1;
+	}
 
       p += len;
       if (p[-1] != '\n')
@@ -1691,6 +1762,9 @@ readline (linebuffer, stream, filename, lineno)
 	}
     }
 
+  if (ferror (stream))
+    pfatal_with_name (filename);
+
   return nlines;
 }
 
@@ -1727,7 +1801,7 @@ construct_include_path (arg_dirs)
 	      dir = expanded;
 	  }
 
-	if (stat (dir, &stbuf) == 0 && S_ISDIR (stbuf.st_mode))
+	if (safe_stat (dir, &stbuf) == 0 && S_ISDIR (stbuf.st_mode))
 	  {
 	    if (idx == max - 1)
 	      {
@@ -1744,7 +1818,7 @@ construct_include_path (arg_dirs)
   /* Now add at the end the standard default dirs.  */
 
   for (i = 0; default_include_directories[i] != 0; ++i)
-    if (stat (default_include_directories[i], &stbuf) == 0
+    if (safe_stat (default_include_directories[i], &stbuf) == 0
 	&& S_ISDIR (stbuf.st_mode))
       dirs[idx++] = default_include_directories[i];
 
